"""Core API for Buffer, Physics, Task, Environment and wrappers."""
from .control import Buffer, Physics, Task, Environment
from . import wrappers

__all__ = ('Buffer', 'Physics', 'Task', 'Environment', 'wrappers')



from typing import Optional, Tuple, Union

import gym
from gym.spaces import Discrete
import torch


class AsTensor(gym.Wrapper):
    """Wrapper that transforms data types to PyTorch tensors."""

    def __init__(
        self,
        env: gym.Env,
        dtype: torch.dtype = torch.float32,
        device: Optional[Union[str, torch.device]] = None,
    ) -> None:
        super().__init__(env)
        self.dtype = dtype
        self.device = device

    @property
    def state_dim(self) -> int:
        return self.observation_space.shape[0]

    @property
    def action_dim(self) -> int:
        if isinstance(self.action_space, Discrete):
            return self.action_space.n
        else:
            return self.action_space.shape[0]

    @property
    def action_sample(self) -> torch.Tensor:
        action = self.action_space.sample()
        return self.as_tensor(action, self.dtype, self.device)

    def reset(self) -> torch.Tensor:
        observation = self.env.reset()
        return self.as_tensor(observation, self.dtype, self.device)

    def step(self, action: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, dict]:
        if isinstance(self.action_space, Discrete):
            action = action.cpu().detach().item()
        else:
            action = action.view(-1).cpu().detach().numpy()
        observation, reward, done, info = self.env.step(action)
        observation = self.as_tensor(observation, self.dtype, self.device)
        reward = self.as_tensor(reward, self.dtype, self.device)
        done = self.as_tensor(done, self.dtype, self.device)
        return observation, reward, done, info

    @staticmethod
    def as_tensor(data, dtype, device):
        return torch.as_tensor(data, dtype=dtype, device=device).view(1, -1)



"""A gym.Env subclass for buffered control-specific environments based on physics and task."""
from typing import Callable, Optional, Tuple, Union, NamedTuple
import abc
import contextlib

import numpy as np
import gym
from gym.utils import seeding


class Timestep(NamedTuple):

    timestep: int
    time: float
    control: Union[float, np.ndarray]
    physical_control: Union[float, np.ndarray]
    state: np.ndarray
    action: Union[int, np.ndarray]
    observation: np.ndarray
    reference: np.ndarray
    reward: float


class Trajectory(Timestep):
    pass


class Buffer(abc.ABC):
    """Buffer for environments."""

    _data = []

    def __len__(self) -> int:
        return len(self._data)

    def __copy__(self):
        # TODO: handle copy in the future
        raise NotImplementedError("Call the method 'copy()' instead")

    @property
    def timestep(self) -> Optional[Timestep]:
        """Returns the current timestep if the buffer is not empty."""
        if len(self._data) == 0:
            return None
        return self._data[-1]

    @property
    def trajectory(self) -> Optional[Trajectory]:
        """Returns the current trajector if the buffer is not empty."""
        if len(self._data) == 0:
            return None
        return Trajectory(*zip(*self._data))

    def copy(self):
        "Returns a copy of the buffer."
        import copy
        instance = self.__class__.__new__(self.__class__)
        instance.__dict__['_data'] = copy.deepcopy(self.__dict__['_data'])
        return instance

    def empty(self) -> bool:
        "Whether the buffer is empty."
        return len(self._data) == 0

    def push(self, *timestep) -> None:
        "Push a timestep into the buffer."
        self._data.append(Timestep(*timestep))

    def flush(self) -> None:
        "Flushes the buffer."
        self._data.clear()


class Physics(abc.ABC):
    """Simulates a controlled physical dynamical system."""

    _timestep = None
    _time = None
    _control = None
    _physical_control = None
    _state = None

    _np_random = None

    @property
    def np_random(self) -> seeding.RandomNumberGenerator:
        """Lazily seeds for the random number generator(s) since it is expensive and only needed
        if sampling.
        """
        if self._np_random is None:
            self.seed()
        return self._np_random

    @abc.abstractmethod
    def dynamics(
        self,
        time: float,
        state: np.ndarray,
        control: Union[float, np.ndarray],
    ) -> np.ndarray:
        """Defines the ODEs for the dynamical system at every call.

        Should be overridden by all subclasses.

        Args:
            time (float): The simulation time (implicit in the ODEs).
            state (np.ndarray): The simulation state.
            control (Union[float, np.ndarray]): The control signal for the actuators.
        """
        raise NotImplementedError

    def set_control(self, control: Union[float, np.ndarray]) -> None:
        """Sets the control signal for the actuators.

        ```python
        self._control = control
        self._physical_control = control
        ```

        Args:
            control (Union[float, np.ndarray]): A valid control signal.
        """
        raise NotImplementedError

    @contextlib.contextmanager
    def reset_context(self):
        """Context manager for resetting the simulation state.

        Sets the internal simulation to a default state when entering the block.

        ```python
        with physics.reset_context():
            task.reset(physics)
        task.step()
        ```

        Yields:
            The `Physics` instance.
        """
        try:
            self.reset()
        except PhysicsError:
            pass
        finally:
            yield self

    @abc.abstractmethod
    def reset(self) -> None:
        """Resets the simulation state."""
        raise NotImplementedError

    @abc.abstractmethod
    def step(self, sampling_rate: float = 1.) -> None:
        """Updates the simulation state.

        Args:
            sampling_rate (float): How often to repeatedly update the simulation state (optional).
        """
        raise NotImplementedError

    def seed(self, seed: Optional[int] = None) -> None:
        """Seeds for the random number generator(s)."""
        self._np_random, _ = seeding.np_random(seed)

    def timestep(self) -> int:
        """Returns the simulation timestep."""
        return self._timestep

    def time(self) -> float:
        """Returns the elapsed simulation time."""
        return self._time

    def control(self) -> Union[float, np.ndarray]:
        """Returns the simulating control signal for the actuators."""
        return self._control

    def physical_control(self) -> Union[float, np.ndarray]:
        """Returns the physical control signal for the actuators."""
        return self._physical_control

    def state(self) -> np.ndarray:
        """Returns the simulation state."""
        return self._state


class PhysicsError(RuntimeError):
    """Raised if the state of the physical simulation becomes divergent."""
    pass


class Task(abc.ABC):
    """Defines a task based on the physical simulation."""

    registered_references = {}
    registered_rewards = {}

    _observation = None
    _reference = None
    _reward = None

    _np_random = None

    @property
    def np_random(self) -> seeding.RandomNumberGenerator:
        """Lazily seeds since it is expensive and only needed if sampling."""
        if self._np_random is None:
            self.seed()
        return self._np_random

    @abc.abstractmethod
    def target(self, time: float) -> np.ndarray:
        """Defines the target for the task at every call.

        Should be overridden by all subclasses.

        Args:
            time (float): The simulation time.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def action_space(self, physics: Physics) -> gym.Space:
        """Returns a valid action space for the task."""
        raise NotImplementedError

    @abc.abstractmethod
    def observation_space(self, physics: Physics) -> gym.Space:
        """Returns a valid observation space for the task."""
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self, physics: Physics) -> None:
        """Sets the initial state of the environment at the start of an episode.

        Args:
            physics (Physics): Instance of `Physics`.
        """
        raise NotImplementedError

    def after_reset(self, physics: Physics) -> None:
        """Optional method to update the task before stepping forward."""
        pass

    def before_step(self, action: Union[int, np.ndarray], physics: Physics) -> None:
        """Updates the task from the provided action.

        Should be called before stepping the physical simulation.

        ```python
        control = action
        physics.set_control(control)
        ```

        Args:
            action (Union[int, np.ndarray]): A valid action drawing from the action space.
            physics (Physics): Instance of `Physics`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def step(self, physics: Physics) -> None:
        """Updates the task by stepping the physical simulation.

        Args:
            physics (Physics): Instance of `Physics`.
        """
        raise NotImplementedError

    def after_step(self, physics: Physics) -> None:
        """Optional method to update the task after stepping the physical simulation."""
        pass

    def seed(self, seed: Optional[int] = None) -> None:
        """Seeds for the random number generator(s)."""
        self._np_random, _ = seeding.np_random(seed)

    @classmethod
    def register_reference(cls, reference: Callable) -> Callable:
        """Decorator to register a reference computation function.

        ```python
        @Task.register_reference
        def const(t: np.ndarray) -> np.ndarray:
            return CONST + np.zeros_like(t)
        ```

        Args:
            reference (Callable): The reference computation function.
        """
        cls.registered_references[reference.__name__] = reference
        return reference

    @classmethod
    def register_reward(cls, reward: Callable) -> Callable:
        """Decorator to register a reward computation function.

        ```python
        @Task.register_reward
        def in_tolerance(
            achieved: Union[float, np.ndarray],
            desired: Union[float, np.ndarray],
            tolerance: float,
        ) -> float:
            return float(abs(achieved - desired) / desired < tolerance)
        ```

        Args:
            reward (Callable): The reward computation function.
        """
        cls.registered_rewards[reward.__name__] = reward
        return reward

    def observation(self, physics: Physics) -> np.ndarray:
        """Returns an observation from the environment."""
        raise NotImplementedError

    def reference(self, physcis: Physics) -> np.ndarray:
        """Returns the reference corresponding to the task target."""
        raise NotImplementedError

    def reward(self, physics: Physics) -> float:
        """Returns a reward from the environment."""
        raise NotImplementedError

    def terminated(self, physics: Physics) -> bool:
        """Whether the episode should be terminated."""
        return False

    def info(self, physics: Physics) -> dict:
        """Returns the extras."""
        return {}


class Environment(gym.Env):
    """Buffered environment based on physics and task."""

    _buffer = None

    def __init__(self, physics: Physics, task: Task) -> None:
        """Initializes an environment.

        Args:
            physics: Instance of `Physics`.
            task: Instance of `Task`.
        """
        self._buffer = Buffer()
        self._physics = physics
        self._task = task
        self.action_space = self._task.action_space(self._physics)
        self.observation_space = self._task.observation_space(self._physics)

    @property
    def buffer(self) -> Buffer:
        """Returns the buffer of the environment."""
        return self._buffer

    @property
    def physics(self) -> Physics:
        """Returns the physics of the environment."""
        return self._physics

    @property
    def task(self) -> Task:
        """Returns the task of the environment."""
        return self._task

    def reset(self) -> np.ndarray:
        """Resets the environment starting a new episode.

        Returns:
            observation (np.ndarray): An observation from the environment.
        """
        self._buffer.flush()
        with self._physics.reset_context():
            self._task.reset(self._physics)
        self._task.after_reset(self._physics)
        timestep = self._physics.timestep()
        time = self._physics.time()
        control = None
        physical_control = None
        state = self._physics.state()
        action = None
        observation = self._task.observation(self._physics)
        reference = self._task.reference(self._physics)
        reward = None
        self._buffer.push(
            timestep,
            time,
            control,
            physical_control,
            state,
            action,
            observation,
            reference,
            reward,
        )
        return observation

    def step(self, action: Union[int, np.ndarray]) -> Tuple[np.ndarray, float, bool, dict]:
        """Updates the environment.

        Args:
            action (Union[int, np.ndarray]):

        Returns:
            observation (np.ndarray): An observation from the environment.
            reward (float): An observation from the environment.
            terminated (bool): Whether the episode should be terminated.
            info (dict): The extras.
        """
        if self._buffer.empty():
            raise RuntimeError
        self._task.before_step(action, self._physics)
        self._task.step(self._physics)
        self._task.after_step(self._physics)
        timestep = self._physics.timestep()
        time = self._physics.time()
        control = self._physics.control()
        physical_control = self._physics.physical_control()
        state = self._physics.state()
        observation = self._task.observation(self._physics)
        reference = self._task.reference(self._physics)
        reward = self._task.reward(self._physics)
        terminated = self._task.terminated(self._physics)
        info = self._task.info(self._physics)
        self._buffer.push(
            timestep,
            time,
            control,
            physical_control,
            state,
            action,
            observation,
            reference,
            reward,
        )
        return observation, reward, terminated, info

    def close(self) -> None:
        """Cleans up the buffer of the environment."""
        self._buffer = None

    def seed(self, seed: Optional[int] = None) -> None:
        """Seeds for the random number generator(s)."""
        self._np_random, _ = seeding.np_random(seed)
        self._physics.seed(seed)
        self._task.seed(seed)



"""The essential wrappers for the control-specific environments."""
from typing import Callable, Tuple, Union
import pathlib

import numpy as np
import gym
from gym.spaces import Box
import imageio


class TransformObservation(gym.ObservationWrapper):
    """Wrapper that can transform observation via `transform()` function.

    Warning:
        Observation space should be updated if necessary.
    """

    def __init__(self, env: gym.Env, transform: Callable) -> None:
        super().__init__(env)
        self.transform = transform

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return self.transform(observation)


class TransformAction(gym.ActionWrapper):
    """Wrapper that can transform action via `transform()` function.

    Warning:
        Action space should be updated if necessary.
    """

    def __init__(self, env: gym.Env, transform: Callable) -> None:
        super().__init__(env)
        self.transform = transform

    def action(self, action: Union[int, np.ndarray]) -> Union[int, np.ndarray]:
        return self.transform(action)


class TransformReward(gym.RewardWrapper):
    """Wrapper that can transform reward via `transform()` function.

    Warning:
        Reward range should be updated if necessary.
    """

    def __init__(self, env: gym.Env, transform: Callable) -> None:
        super().__init__(env)
        self.transform = transform

    def reward(self, reward: float) -> float:
        return self.transform(reward)


class LimitTimestep(gym.Wrapper):
    """Wrapper that limits timesteps per episode."""

    def __init__(self, env: gym.Env, max_episode_steps: int) -> None:
        super().__init__(env)
        self.max_episode_steps = max_episode_steps

    def step(self, action: Union[int, np.ndarray]) -> Tuple[np.ndarray, float, bool, dict]:
        observation, reward, terminated, info = super().step(action)
        if len(self.buffer) > self.max_episode_steps:
            info['truncated'] = not terminated
            terminated = True
        return observation, reward, terminated, info


class TrackEpisode(gym.Wrapper):
    """Wrapper that keeps track of cumulative reward and episode length."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.episode_return = None
        self.episode_length = None

    def reset(self) -> np.ndarray:
        observation = super().reset()
        self.episode_return = 0.0
        self.episode_length = 0
        return observation

    def step(self, action: Union[int, np.ndarray]) -> Tuple[np.ndarray, float, bool, dict]:
        observation, reward, terminated, info = super().step(action)
        self.episode_return += reward
        self.episode_length += 1
        if terminated:
            info['episode_return'] = self.episode_return
            info['episode_length'] = self.episode_length
        return observation, reward, terminated, info


class RecordEpisode(gym.Wrapper):
    """Wrapper that records video of episode."""

    def __init__(self, env: gym.Env, path: Union[str, pathlib.Path], **render_kwargs) -> None:
        super().__init__(env)
        self.path = pathlib.Path(path)
        self.path.mkdir(parents=True, exist_ok=True)
        self.render_kwargs = render_kwargs
        self._frames = []

    def step(self, action: Union[int, np.ndarray]) -> Tuple[np.ndarray, float, bool, dict]:
        observation, reward, terminated, info = super().step(action)
        fig = super().render(**self.render_kwargs)
        frame = self.path.joinpath(f'fig{self.buffer.timestep.timestep}.png')
        fig.savefig(frame)
        self._frames.append(frame)
        if terminated:
            with imageio.get_writer('episode.gif', mode='I') as writer:
                for frame in self._frames:
                    writer.append_data(imageio.imread(frame))
        return observation, reward, terminated, info


class FullObservation(gym.ObservationWrapper):
    """Wrapper that observes physical internal state."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=self.physics.state_min,
            high=self.physics.state_max,
            dtype=self.physics.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return self.buffer.timestep.state


class TimeAwareObservation(gym.ObservationWrapper):
    """Wrapper that augments the observation with current time."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, np.inf),
            dtype=self.observation_space.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return np.append(observation,
                         self.buffer.timestep.time).astype(self.observation_space.dtype)


class TimestepAwareObservation(gym.ObservationWrapper):
    """Wrapper that augments the observation with current timestep."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, np.inf),
            dtype=self.observation_space.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return np.append(observation,
                         self.buffer.timestep.timestep).astype(self.observation_space.dtype)


class ReferenceAwareObservation(gym.ObservationWrapper):
    """Wrapper that augments the observation with current reference."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, np.inf),
            dtype=self.observation_space.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return np.append(observation,
                         self.buffer.timestep.reference).astype(self.observation_space.dtype)


class ToleranceAwareObservation(gym.ObservationWrapper):
    """Wrapper that augments the observation with current tolerance."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, np.inf),
            dtype=self.observation_space.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        reference = self.buffer.timestep.reference
        tolerance = self.task.tolerance
        _in_tolerance = self.in_tolerance(observation, reference, tolerance)
        return np.append(observation, _in_tolerance).astype(self.observation_space.dtype)

    @staticmethod
    def in_tolerance(achieved, desired, tolerance):
        return float(np.abs(achieved - desired) / desired < tolerance)


class ActionAwareObservation(gym.ObservationWrapper):
    """Wrapper that augments the observation with current action."""

    def __init__(self, env: gym.Env) -> None:
        super().__init__(env)
        self.observation_space = Box(
            low=np.append(self.observation_space.low, 0.0),
            high=np.append(self.observation_space.high, np.inf),
            dtype=self.observation_space.dtype,
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return np.append(observation,
                         self.buffer.timestep.action).astype(self.observation_space.dtype)


class RescaleAction(gym.ActionWrapper):
    """Wrapper that affinely rescales continuous action space."""

    def __init__(
        self,
        env: gym.Env,
        action_min: Union[float, np.ndarray] = 0.0,
        action_max: Union[float, np.ndarray] = 1.0,
    ) -> None:
        super().__init__(env)
        self._from_range = (self.action_space.low, self.action_space.high)
        self.action_space = Box(
            low=action_min,
            high=action_max,
            shape=(1,),
            dtype=self.action_space.dtype,
        )
        self._to_range = (self.action_space.low, self.action_space.high)

    def action(self, action: Union[int, np.ndarray]) -> Union[int, np.ndarray]:
        raise self.rescale(action, self._from_range, self._to_range)

    @staticmethod
    def rescale(data, from_range, to_range):
        _low, _high = from_range
        low, high = to_range
        return low + (data - _low) * (high - low) / (_high - _low)



"""Reinforcement learning environments for CRNs."""
from typing import Callable, Optional, Union, Type
import pathlib

import numpy as np
import scipy.integrate as integrate
import scipy.signal as signal
import gym
from gym.spaces import Discrete, Box

from .control import Buffer, Physics, Task, Environment, wrappers

# Reference registry for tracking task


@Task.register_reference
def const(t, scale):
    "Constant wave."
    return scale + np.zeros_like(t)


@Task.register_reference
def square(t, scale, amplitude, period, phase):
    "Square wave."
    return scale + amplitude * signal.square(2 * np.pi * t / period + phase).astype(t.dtype)


@Task.register_reference
def sine(t, scale, amplitude, period, phase):
    "Sine (or Cosine) wave."
    return scale + amplitude * np.sin(2 * np.pi * t / period + phase).astype(t.dtype)


@Task.register_reference
def bpf(t, switches):
    "Band-pass filter (BPF)."
    y = np.zeros_like(t)
    mask_nan = True
    for i in range(switches.shape[0]):
        mask = (t == switches[i, 0])
        np.place(y, mask, switches[i, 1])
        mask_nan &= (1 - mask)
    np.place(y, mask_nan, np.nan)
    return y


# Reward registry for tracking task


@Task.register_reward
def inverse_ae(achieved, desired, tolerance, n=1.0):
    "Inverse of absolute error (AE)."
    return float(np.abs(achieved - desired)**(-n))


@Task.register_reward
def negative_ae(achieved, desired, tolerance, n=1.0):
    "Negative absolute error (AE)."
    return float(-np.abs(achieved - desired)**n)


@Task.register_reward
def negative_re(achieved, desired, tolerance):
    "Negative relative error (RE)."
    return float(-np.abs(achieved - desired) / desired)


@Task.register_reward
def in_tolerance(achieved, desired, tolerance):
    "Whether falling within tolerance."
    return float(np.abs(achieved - desired) / desired < tolerance)


@Task.register_reward
def gauss(achieved, desired, tolerance):
    "Gauss."
    return float(np.exp(-0.5 * (achieved - desired)**2 / tolerance**2))


@Task.register_reward
def scaled_combination(achieved, desired, tolerance, a=100.0, b=10.0):
    "Scaled combination of errors."
    return negative_ae(achieved, desired, tolerance) * a \
        + in_tolerance(achieved, desired, tolerance) * b


def make(cls: str, kwargs: dict):
    physics = CRN(**kwargs['physics'])
    if cls == 'CRN-v0':
        task = DiscreteTrack(**kwargs['task'])
    elif cls == 'CRNContinuous-v0':
        task = Track(**kwargs['task'])
    else:
        raise RuntimeError
    env = CRNEnv(physics, task, **kwargs['environment'])
    return CRNWrapper(env, **kwargs['wrappers'])


class CRN(Physics):

    def __init__(
        self,
        init_state: np.ndarray,
        integrator: str = 'RK45',
        n_sub_timesteps: int = 100,
        system_noise: float = 0.0,
        actuation_noise: float = 0.0,
        state_min: Union[float, np.ndarray] = 0.0,
        state_max: Union[float, np.ndarray] = float(np.finfo(np.float32).max),
        state_dtype: Type = np.float32,
        state_info: dict = {},
        control_min: float = 0.0,
        control_max: float = 1.0,
        control_dtype: Type = float,
        control_info: dict = {},
        ode: Optional[Callable] = None,
        **ode_kwargs,
    ) -> None:
        self.ode = ode
        self.ode_kwargs = ode_kwargs
        self.init_state = init_state
        self.integrator = integrator
        self.n_sub_timesteps = n_sub_timesteps
        self.system_noise = system_noise
        self.actuation_noise = actuation_noise
        self.state_shape = init_state.shape
        self.state_min = state_min.astype(state_dtype) if isinstance(
            state_min, np.ndarray) else np.full(self.state_shape, state_min, dtype=state_dtype)
        self.state_max = state_max.astype(state_dtype) if isinstance(
            state_max, np.ndarray) else np.full(self.state_shape, state_max, dtype=state_dtype)
        self.state_dtype = state_dtype
        self.state_info = state_info
        self.control_min = control_min
        self.control_max = control_max
        self.control_dtype = control_dtype
        self.control_info = control_info
        self._timestep = 0
        self._time = 0.0
        self._state = self.init_state

    def dynamics(self, time: float, state: np.ndarray, control: float):
        if self.ode is None:
            raise NotImplementedError
        return self.ode(state, control, **self.ode_kwargs) \
            + self.np_random.normal(0.0, self.system_noise)

    def reset(self) -> None:
        self._timestep = 0
        self._time = 0.0
        self._state = self.init_state
        self._control = None
        self._physical_control = None

    def set_control(self, control: float) -> None:
        self._control = control
        control += self.np_random.normal(0.0, self.actuation_noise)
        control = min(max(control, self.control_min), self.control_max)
        self._physical_control = control

    def step(self, sampling_rate: float) -> None:
        delta = sampling_rate / self.n_sub_timesteps
        sol = integrate.solve_ivp(
            self.dynamics,
            (0, sampling_rate),
            self._state,
            method=self.integrator,
            t_eval=np.arange(0, sampling_rate + delta, delta),
            args=(self._physical_control,),
        )
        self._state = sol.y[:, -1]
        self._state = np.clip(self._state, self.state_min, self.state_max)
        self._timestep += 1
        self._time += sampling_rate

    def state(self) -> np.ndarray:
        return self._state.astype(self.state_dtype)


class Track(Task):

    def __init__(
        self,
        sampling_rate: float,
        dim_observed: int,
        tolerance: float,
        reward: Union[str, Callable],
        reward_kwargs: dict = {},
        reward_info: dict = {},
        observation_noise: float = 0.0,
        action_min: Union[float, np.ndarray] = -1.0,
        action_max: Union[float, np.ndarray] = 1.0,
        action_dtype: Type = np.float32,
        action_info: dict = {},
        tracking: Optional[Union[str, Callable]] = None,
        **tracking_kwargs,
    ) -> None:
        self.tracking = self.registered_references.get(tracking, None) if isinstance(
            tracking, str) else tracking
        self.tracking_kwargs = tracking_kwargs
        self.sampling_rate = sampling_rate
        self.dim_observed = dim_observed
        self.tolerance = tolerance
        self.reward_func = reward if callable(reward) else self.registered_rewards[reward]
        self.reward_kwargs = reward_kwargs
        self.reward_info = reward_info
        self.observation_noise = observation_noise
        self.action_shape = (1,)
        self.action_min = action_min.astype(action_dtype) if isinstance(
            action_min, np.ndarray) else np.full(self.action_shape, action_min, dtype=action_dtype)
        self.action_max = action_max.astype(action_dtype) if isinstance(
            action_max, np.ndarray) else np.full(self.action_shape, action_max, dtype=action_dtype)
        self.action_dtype = action_dtype
        self.action_info = action_info

    def target(self, time: np.ndarray):
        if self.tracking is None:
            raise NotImplementedError
        return self.tracking(time, **self.tracking_kwargs)

    def action_space(self, physics: Physics) -> Box:
        return Box(
            low=self.action_min,
            high=self.action_max,
            dtype=self.action_dtype,
        )

    def observation_space(self, physics: Physics) -> Box:
        return Box(
            low=physics.state_min[[self.dim_observed]],
            high=physics.state_max[[self.dim_observed]],
            dtype=physics.state_dtype,
        )

    def reset(self, physics: Physics) -> None:
        self._observation = None
        self._reference = None
        self._reward = None

    def before_step(self, action: Union[int, np.ndarray], physics: Physics) -> None:
        if isinstance(self.action_space(physics), Discrete):
            action = (action + 1) / self.action_space(physics).n
        else:
            action = float(action[0])
            action = (action - float(self.action_min)) / (float(self.action_max) -
                                                          float(self.action_min))
        control = physics.control_min + action * (physics.control_max - physics.control_min)
        physics.set_control(control)

    def step(self, physics: Physics) -> None:
        physics.step(self.sampling_rate)

    def reference(self, physics: Physics) -> np.ndarray:
        time = np.array([physics.time()]).astype(physics.state_dtype)
        self._reference = self.target(time)
        return self._reference.astype(physics.state_dtype)

    def observation(self, physics: Physics) -> np.ndarray:
        state = physics.state()
        self._observation = state[[self.dim_observed]]
        self._observation += self.np_random.normal(0.0, self.observation_noise)
        self._observation = np.clip(self._observation, physics.state_min[[self.dim_observed]],
                                    physics.state_max[[self.dim_observed]])
        return self._observation.astype(physics.state_dtype)

    def reward(self, physics: Physics) -> float:
        self._reward = self.reward_func(self._observation, self._reference, self.tolerance,
                                        **self.reward_kwargs)
        return self._reward


class DiscreteTrack(Track):

    def action_space(self, physics: Physics) -> Discrete:
        return Discrete(n=20)


class CRNEnv(Environment):

    def __init__(self, physics: Physics, task: Task) -> None:
        super().__init__(physics, task)

    def render(self, mode: str = 'human', buffer: Optional[Buffer] = None):
        if self._buffer.empty() and (buffer is None):
            raise RuntimeError
        buffer = self._buffer if buffer is None else buffer
        tolerance = self._task.tolerance
        sampling_rate = self._task.sampling_rate
        dim_observed = self._task.dim_observed
        # Data: reference trajectory & state / observation  vs. time
        time = np.array(buffer.trajectory.time)
        state = np.stack(buffer.trajectory.state, axis=1)
        observation = np.concatenate(buffer.trajectory.observation, axis=0)
        delta = 0.1     # simulation sampling rate
        time_reference = np.arange(0, time[-1] + delta, delta)
        reference = self.task.target(time_reference)
        # Data: control signal vs. time
        time_control = np.concatenate([
            np.arange(sampling_rate * i, sampling_rate * (i + 2), sampling_rate)
            for i in range(len(buffer) - 1)
        ])
        control = np.array(buffer.trajectory.control[1:]).repeat(2)
        physical_control = np.array(buffer.trajectory.physical_control[1:]).repeat(2)
        # Data: reward vs. time
        time_reward = time[1:]
        reward = np.array(buffer.trajectory.reward[1:])
        # Info: reference trajectory & state / observation  vs. time
        state_info = self._physics.state_info
        observation_info = {
            'color': state_info['color'][dim_observed],
            'label': state_info['label'][dim_observed],
            'ylim': state_info['ylim'],
        }
        # Info: control signal vs. time
        control_info = self._physics.control_info
        # Info: reward vs. time
        reward_info = self._task.reward_info
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            pass

        try:
            import seaborn as sns
            sns.set_theme(style='darkgrid')
        except ImportError:
            pass
        # Partially shown
        if mode == 'human':
            fig, axes = plt.subplots(
                nrows=2,
                ncols=1,
                figsize=(7, 5),
                sharex=True,
                gridspec_kw={'height_ratios': [2, 1]},
            )
            fig.tight_layout()
            # Subplot: reference trajectory & observation vs. time
            self.plot_reference(axes[0], time_reference, reference, tolerance)
            self.plot_observation(axes[0], time, observation, state=None, **observation_info)
            # Subplot: control signal vs. time
            self.plot_control(axes[1],
                              time_control,
                              control,
                              physical_control=None,
                              **control_info)
            axes[1].set_xlabel('Time (min)')
            plt.close()
        # Fully shown
        else:
            fig, axes = plt.subplots(nrows=2,
                                     ncols=2,
                                     figsize=(10, 5),
                                     sharex=True,
                                     gridspec_kw={'height_ratios': [2, 1]})
            fig.tight_layout()
            # Subplot: reference trajectory & state vs. time
            self.plot_reference(axes[0, 0], time_reference, reference, tolerance)
            self.plot_state(axes[0, 0], time, state, **state_info)
            # Subplot: control signal vs. time
            self.plot_control(axes[1, 0], time_control, control, physical_control, **control_info)
            axes[1, 0].set_xlabel('Time (min)')
            # Subplot: reference trajectory & observation vs. time
            self.plot_reference(axes[0, 1], time_reference, reference, tolerance)
            self.plot_observation(axes[0, 1], time, observation, state[dim_observed],
                                  **observation_info)
            # Subplot: reward vs. time
            self.plot_reward(axes[1, 1], time_reward, reward, **reward_info)
            axes[1, 1].set_xlabel('Time (min)')
            plt.close()
        return fig

    @staticmethod
    def plot_state(ax, time, state, color, label, ylim):
        for i in range(state.shape[0]):
            ax.plot(time, state[i], '-', color=color[i], label=label[i], alpha=0.85)
            if len(time) > 0:
                ax.plot(time[-1], state[i][-1], marker='.', color=color[i])
        ax.legend(loc='upper right', framealpha=0.2)
        ax.set_ylim(ylim)
        ax.set_ylabel('')

    @staticmethod
    def plot_control(ax, time, control, physical_control, color, label, ylim):
        ax.plot(time, control, '-', color=color, label=label, alpha=0.85)
        if len(time) > 0:
            ax.plot(time[-1], control[-1], marker='.', color=color)
        if physical_control is not None:
            ax.plot(time,
                    physical_control,
                    '--',
                    color=color,
                    label=label + ' performed',
                    alpha=0.35)
            if len(time) > 0:
                ax.plot(time[-1], physical_control[-1], marker='.', color=color, alpha=0.5)
        ax.legend(loc='upper right', framealpha=0.2)
        ax.set_ylim(ylim)
        ax.set_ylabel('')

    @staticmethod
    def plot_reference(ax, time, reference, tolerance, color='grey'):
        if np.isnan(reference).any():
            ax.plot(time, reference, 'x-', color=color)
            # ax.scatter(time, reference, color=color)
            ax.errorbar(time, reference, yerr=reference * tolerance, color=color)
        else:
            ax.plot(time, reference, '--', color=color)
            ax.fill_between(time,
                            reference * (1 - tolerance),
                            reference * (1 + tolerance),
                            color=color,
                            alpha=0.15)
        ax.set_ylabel('')

    @staticmethod
    def plot_observation(ax, time, observation, state, color, label, ylim):
        ax.plot(time, observation, '-', color=color, label=label + ' observed', alpha=0.85)
        if len(time) > 0:
            ax.plot(time[-1], observation[-1], marker='.', color=color)
        if state is not None:
            ax.plot(time, state, '--', color=color, label=label, alpha=0.35)
            if len(time) > 0:
                ax.plot(time[-1], state[-1], marker='.', color=color, alpha=0.5)
        ax.legend(loc='upper right', framealpha=0.2)
        ax.set_ylim(ylim)
        ax.set_ylabel('')

    @staticmethod
    def plot_reward(ax, time, reward, color, label, ylim):
        ax.plot(time, reward, color=color, label=label + ' reward', alpha=0.85)
        if len(time) > 0:
            ax.plot(time[-1], reward[-1], marker='.', color=color)
        ax.legend(loc='upper right', framealpha=0.2)
        ax.set_ylim(ylim)
        ax.set_ylabel('')


class CRNWrapper(gym.Wrapper):

    def __init__(
        self,
        env: CRNEnv,
        max_episode_steps: Optional[int] = None,
        full_observation: bool = False,
        time_aware: bool = False,
        timestep_aware: bool = False,
        reference_aware: bool = False,
        tolerance_aware: bool = False,
        action_aware: bool = False,
        rescale_action: bool = False,
        action_min: Union[float, np.ndarray] = 0.0,
        action_max: Union[float, np.ndarray] = 1.0,
        track_episode: bool = False,
        record_episode: bool = False,
        path: Union[str, pathlib.Path] = 'episode',
        **render_kwargs,
    ):
        if max_episode_steps is not None:
            env = wrappers.LimitTimestep(env, max_episode_steps)
        if full_observation:
            env = wrappers.FullObservation(env)
        if time_aware:
            env = wrappers.TimeAwareObservation(env)
        if timestep_aware:
            env = wrappers.TimestepAwareObservation(env)
        if reference_aware:
            env = wrappers.ReferenceAwareObservation(env)
        if tolerance_aware:
            env = wrappers.ToleranceAwareObservation(env)
        if action_aware:
            env = wrappers.ActionAwareObservation(env)
        if rescale_action:
            env = wrappers.RescaleAction(env, action_min, action_max)
        if track_episode:
            env = wrappers.TrackEpisode(env)
        if record_episode:
            env = wrappers.RecordEpisode(env, path, **render_kwargs)
        super().__init__(env)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cybergenetics',
 'cybergenetics.envs',
 'cybergenetics.envs.control',
 'cybergenetics.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['gym>=0.25.0,<0.26.0',
 'imageio>=2.20.0,<3.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'torch>=1.12.0,<2.0.0']

setup_kwargs = {
    'name': 'cybergenetics',
    'version': '0.0.2',
    'description': 'Cybergenetics is a controlled simulating environment for chemical reaction networks (CRNs) and co-cultures.',
    'long_description': None,
    'author': 'Yi Zhang, Quentin Badolle',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
